#include <vector>
#include <map>
#include <string>

class cJavaClass;
class cJVM;
struct METHOD_INFO;
struct FIELD_INFO;
struct STACK_ITEM;

//-----------------------------------------------------------------------------
// Name: class cVirtualMachine
// Desc: Reprezents virtual machine.
//-----------------------------------------------------------------------------
class cJVM
{
private:
	//typedef stl map for easier use
	typedef std::map<std::string,cJavaClass*>				ClassMap;
	typedef std::map<std::string,cJavaClass*>::iterator		ClassMapItor;
	typedef std::map<std::string,cJavaClass*>::value_type	ClassMapValue;

	//map of all loaded classes
	ClassMap	m_Classes;

	//Array of pointers to instruction functions
	void (cJVM::*Instructions[INST_NUMINST])();

	//program counter
	u1	*m_pPC;
	//stack
	cStack			m_Stack;
	//heap
	cHeap			m_Heap;
	//class of current method
	cJavaClass		*m_pCurClass;
	//flag tells if an error occured
	bool			m_bError;

private:
	//Read u2 (big endian)
	u2 ReadU2();
	//Get double from high, low bytes
	void GetDouble(double &d,u4 iHighBytes,u4 iLowBytes);
	//Get long from high, low bytes
	void GetLong(long long &l,u4 iHighBytes,u4 iLowBytes);
	//Get high, low bytes from double
	void GetBytesFromDouble(double &d,u4 &iHighBytes,u4 &iLowBytes);
	//Get high, low bytes from long
	void GetBytesFromLong(long long &l,u4 &iHighBytes,u4 &iLowBytes);

	//Breaks execution and prints error msg to stdout
	void Error(char *strError);

	//Instructions
	//new
	void Inst_New();
	//new array
	void Inst_NewArray();
	//anew array
	void Inst_ANewArray();
	//dup
	void Inst_Dup();
	//invoke virtual
	void Inst_InvokeVirtual();
	//invoke special
	void Inst_InvokeSpecial();
	//invoke static
	void Inst_InvokeStatic();
	//aload
	void Inst_ALoad();
	//dload
	void Inst_DLoad();
	//dload 1
	void Inst_DLoad_1();
	//aload 0
	void Inst_ALoad_0();
	//aload 1
	void Inst_ALoad_1();
	//aload 2
	void Inst_ALoad_2();
	//aload 3
	void Inst_ALoad_3();
	//return
	void Inst_Return();
	//ireturn
	void Inst_IReturn();
	//areturn
	void Inst_AReturn();
	//dreturn
	void Inst_DReturn();
	//astore 1
	void Inst_AStore_1();
	//astore 2
	void Inst_AStore_2();
	//astore 3
	void Inst_AStore_3();
	//istore
	void Inst_IStore();
	//istore 1
	void Inst_IStore_1();
	//istore 0
	void Inst_IStore_0();
	//istore 2
	void Inst_IStore_2();
	//istore 3
	void Inst_IStore_3();
	//dstore 1
	void Inst_DStore_1();
	//iconst 0
	void Inst_IConst_0();
	//iconst 1
	void Inst_IConst_1();
	//iconst 2
	void Inst_IConst_2();
	//iconst 3
	void Inst_IConst_3();
	//iconst 4
	void Inst_IConst_4();
	//iconst 5
	void Inst_IConst_5();
	//iload
	void Inst_ILoad();
	//iload 0
	void Inst_ILoad_0();
	//iload 1
	void Inst_ILoad_1();
	//iload 2
	void Inst_ILoad_2();
	//iload 3
	void Inst_ILoad_3();
	//iadd
	void Inst_IAdd();
	//pop
	void Inst_Pop();
	//iastore
	void Inst_IAStore();
	//bipush
	void Inst_BIPush();
	//put static
	void Inst_PutStatic();
	//get static
	void Inst_GetStatic();
	//put field
	void Inst_PutField();
	//get field
	void Inst_GetField();
	//goto
	void Inst_GoTo();
	//baload
	void Inst_BALoad();
	//iinc
	void Inst_IInc();
	//array length
	void Inst_ArrayLength();
	//if i cmp lt
	void Inst_If_ICMPLT();
	//sipush
	void Inst_SIPush();
	//bastore
	void Inst_BAStore();
	//iaload
	void Inst_IALoad();
	//aaload
	void Inst_AALoad();
	//ldc2 w
	void Inst_LDC2_W();
	//dadd
	void Inst_DAdd();
	//d2i
	void Inst_D2I();
	//dconst 0
	void Inst_DConst_0();
	//dconst 1
	void Inst_DConst_1();
	//aastore
	void Inst_AAStore();
	//aconst null
	void Inst_AConst_Null();

	//relocate single object (garbage collector)
	OBJECT* CollectObject(OBJECT *pObj);

public:
	//Constructor
	cJVM();
	//Destructor
	~cJVM();
	
	//Removes all loaded classes
	void FreeAllClasses();
	//Initialize vm. Must be called before anything else
	bool Initialize(u4 iStackSize, u4 iHeapSize);
	//Load class from file and prepare its main method for execution
	bool LoadMain(const char *strFilename);
	//Load class from file
	cJavaClass* LoadClass(const char *strFilename);
	//Run initialization method of suplied class. Called automaticly after class is loaded.
	bool InitializeClass(cJavaClass *pClass);
	//Run method if any was prepared for execution
	bool Run();
	//Run garbage collector
	void RunGC();
};